from django.db import models


class DBMutex(models.Model):
    """
    Models a mutex lock with a ``lock_id`` and a ``creation_time``.

    :type lock_id: str
    :param lock_id: A unique CharField with a max length of 256

    :type creation_time: datetime
    :param creation_time: The creation time of the mutex lock
    """
    lock_id = models.CharField(max_length=256, unique=True)
    creation_time = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = 'db_mutex'



from datetime import timedelta
import functools
import logging

from django.conf import settings
from django.db import transaction, IntegrityError
from django.utils import timezone

from .exceptions import DBMutexError, DBMutexTimeoutError
from .models import DBMutex


LOG = logging.getLogger(__name__)


class db_mutex(object):
    """
    An object that acts as a context manager and a function decorator for acquiring a
    DB mutex lock.
    """
    mutex_ttl_seconds_settings_key = 'DB_MUTEX_TTL_SECONDS'

    def __init__(self, lock_id, suppress_acquisition_exceptions=False):
        """
        This context manager/function decorator can be used in the following way

        .. code-block:: python

            from db_mutex.db_mutex import db_mutex

            # Lock a critical section of code
            try:
                with db_mutex('lock_id'):
                    # Run critical code here
                    pass
            except DBMutexError:
                print('Could not obtain lock')
            except DBMutexTimeoutError:
                print('Task completed but the lock timed out')

            # Lock a function
            @db_mutex('lock_id'):
            def critical_function():
                # Critical code goes here
                pass

            try:
                critical_function()
            except DBMutexError:
                print('Could not obtain lock')
            except DBMutexTimeoutError:
                print('Task completed but the lock timed out')

        :type lock_id: str
        :param lock_id: The ID of the lock one is trying to acquire
        :type suppress_acquisition_exceptions: bool
        :param suppress_acquisition_exceptions: Suppress exceptions when acquiring the lock and instead
            log an error message. Note that this is only applicable when using this as a decorator and
            not a context manager.

        :raises:
            * :class:`DBMutexError <db_mutex.exceptions.DBMutexError>` when the lock cannot be obtained
            * :class:`DBMutexTimeoutError <db_mutex.exceptions.DBMutexTimeoutError>` when the
              lock was deleted during execution

        """
        self.lock_id = lock_id
        self.lock = None
        self.suppress_acquisition_exceptions = suppress_acquisition_exceptions

    def get_mutex_ttl_seconds(self):
        """
        Returns a TTL for mutex locks. It defaults to 30 minutes. If the user specifies None
        as the TTL, locks never expire.

        :rtype: int
        :returns: the mutex's ttl in seconds
        """
        return getattr(settings, self.mutex_ttl_seconds_settings_key, timedelta(minutes=30).total_seconds())

    def delete_expired_locks(self):
        """
        Deletes all expired mutex locks if a ttl is provided.
        """
        ttl_seconds = self.get_mutex_ttl_seconds()
        if ttl_seconds is not None:
            DBMutex.objects.filter(creation_time__lte=timezone.now() - timedelta(seconds=ttl_seconds)).delete()

    def __call__(self, func):
        return self.decorate_callable(func)

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop()

    def start(self):
        """
        Acquires the db mutex lock. Takes the necessary steps to delete any stale locks.
        Throws a DBMutexError if it can't acquire the lock.
        """
        # Delete any expired locks first
        self.delete_expired_locks()
        try:
            with transaction.atomic():
                self.lock = DBMutex.objects.create(lock_id=self.lock_id)
        except IntegrityError:
            raise DBMutexError('Could not acquire lock: {0}'.format(self.lock_id))

    def stop(self):
        """
        Releases the db mutex lock. Throws an error if the lock was released before the function finished.
        """
        if not DBMutex.objects.filter(id=self.lock.id).exists():
            raise DBMutexTimeoutError('Lock {0} expired before function completed'.format(self.lock_id))
        else:
            self.lock.delete()

    def decorate_callable(self, func):
        """
        Decorates a function with the db_mutex decorator by using this class as a context manager around
        it.
        """
        def wrapper(*args, **kwargs):
            try:
                with self:
                    result = func(*args, **kwargs)
                return result
            except DBMutexError as e:
                if self.suppress_acquisition_exceptions:
                    LOG.error(e)
                else:
                    raise e
        functools.update_wrapper(wrapper, func)
        return wrapper



class DBMutexError(Exception):
    """
    Thrown when a lock cannot be acquired.
    """
    pass


class DBMutexTimeoutError(Exception):
    """
    Thrown when a lock times out before it is released.
    """
    pass



from datetime import datetime

from db_mutex.exceptions import DBMutexError, DBMutexTimeoutError
from db_mutex.models import DBMutex
from db_mutex.db_mutex import db_mutex

from django.test import TestCase
from django.test.utils import override_settings
from freezegun import freeze_time


class ContextManagerTestCase(TestCase):
    """
    Tests db_mutex as a context manager.
    """
    @freeze_time('2014-02-01')
    def test_no_lock_before(self):
        """
        Tests that a lock is succesfully acquired.
        """
        # There should be no locks before and after the context manager
        self.assertEqual(DBMutex.objects.count(), 0)
        with db_mutex('lock_id'):
            self.assertEqual(DBMutex.objects.count(), 1)
            m = DBMutex.objects.get(lock_id='lock_id')
            self.assertEqual(m.creation_time, datetime(2014, 2, 1))
        self.assertEqual(DBMutex.objects.count(), 0)

    @freeze_time('2014-02-01')
    def test_lock_before(self):
        """
        Tests when a lock already exists.
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')
        # Try to acquire the lock. It should raise an exception
        with self.assertRaises(DBMutexError):
            with db_mutex('lock_id'):
                raise NotImplementedError
        # The lock should still exist
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())

    @freeze_time('2014-02-01')
    def test_lock_before_suppress_acquisition_errors(self):
        """
        Tests when a lock already exists. Verifies that an exception is thrown when
        suppress_acquisition_errors is True. The exception is still thrown because
        we are using it as a context manager
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')
        # Try to acquire the lock. It should neither acquire nor release it
        with self.assertRaises(DBMutexError):
            with db_mutex('lock_id', suppress_acquisition_exceptions=True):
                raise NotImplementedError
        # The lock should still exist
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())

    @freeze_time('2014-02-01')
    def test_lock_different_id(self):
        """
        Tests that the lock still works even when another lock with a different id exists.
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')
        # Try to acquire the lock with a different ID
        with db_mutex('lock_id2'):
            self.assertEqual(DBMutex.objects.count(), 2)
            m2 = DBMutex.objects.get(lock_id='lock_id2')
            self.assertEqual(m2.creation_time, datetime(2014, 2, 1))
        # The original lock should still exist but the other one should be gone
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())
        self.assertEqual(DBMutex.objects.count(), 1)

    def test_lock_timeout_default(self):
        """
        Tests that the lock timeout works with the default value of 30 minutes.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            orig_lock = DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        with freeze_time('2014-02-01 00:01:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 9 minutes in the future. It should fail
        with freeze_time('2014-02-01 00:09:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 30 minutes in the future. It should pass since the lock timed out
        with freeze_time('2014-02-01 00:30:00'):
            with db_mutex('lock_id'):
                self.assertFalse(DBMutex.objects.filter(id=orig_lock.id).exists())
                self.assertEqual(DBMutex.objects.count(), 1)
                m = DBMutex.objects.get(lock_id='lock_id')
                self.assertEqual(m.creation_time, datetime(2014, 2, 1, 0, 30))

    @override_settings(DB_MUTEX_TTL_SECONDS=None)
    def test_no_lock_timeout(self):
        """
        Tests that the lock timeout works when None is configured as the timeout.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        with freeze_time('2014-02-01 00:01:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 9 minutes in the future. It should fail
        with freeze_time('2014-02-01 00:09:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 30 minutes in the future. It should fail
        with freeze_time('2014-02-01 00:30:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock years in the future. It should fail
        with freeze_time('2016-02-01 00:30:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

    @override_settings(DB_MUTEX_TTL_SECONDS=60 * 60)
    def test_custom_lock_timeout(self):
        """
        Tests that the custom lock timeout works when an hour is configured as the timeout.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            orig_lock = DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        with freeze_time('2014-02-01 00:01:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 31 minutes in the future. It should fail
        with freeze_time('2014-02-01 00:31:00'):
            with self.assertRaises(DBMutexError):
                with db_mutex('lock_id'):
                    raise NotImplementedError

        # Try to acquire the lock 60 minutes in the future. It should pass
        with freeze_time('2014-02-01 01:00:00'):
            with db_mutex('lock_id'):
                self.assertFalse(DBMutex.objects.filter(id=orig_lock.id).exists())
                self.assertEqual(DBMutex.objects.count(), 1)
                m = DBMutex.objects.get(lock_id='lock_id')
                self.assertEqual(m.creation_time, datetime(2014, 2, 1, 1))

    def test_lock_timeout_error(self):
        """
        Tests the case when a lock expires while the context manager is executing.
        """
        with freeze_time('2014-02-01'):
            # Acquire a lock at the given time and release it before it is finished. It
            # should result in an error
            with self.assertRaises(DBMutexTimeoutError):
                with db_mutex('lock_id'):
                    self.assertEqual(DBMutex.objects.count(), 1)
                    m = DBMutex.objects.get(lock_id='lock_id')
                    self.assertEqual(m.creation_time, datetime(2014, 2, 1))

                    # Release the lock before the context manager finishes
                    m.delete()


class FunctionDecoratorTestCase(TestCase):
    """
    Tests db_mutex as a function decorator.
    """
    @freeze_time('2014-02-01')
    def test_no_lock_before(self):
        """
        Tests that a lock is succesfully acquired.
        """
        # There should be no locks before and after the context manager
        self.assertEqual(DBMutex.objects.count(), 0)

        @db_mutex('lock_id')
        def run_get_lock():
            self.assertEqual(DBMutex.objects.count(), 1)
            m = DBMutex.objects.get(lock_id='lock_id')
            self.assertEqual(m.creation_time, datetime(2014, 2, 1))

        run_get_lock()
        self.assertEqual(DBMutex.objects.count(), 0)

    @freeze_time('2014-02-01')
    def test_lock_before(self):
        """
        Tests when a lock already exists.
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')

        @db_mutex('lock_id')
        def run_get_lock():
            raise NotImplementedError

        # Try to acquire the lock. It should raise an exception
        with self.assertRaises(DBMutexError):
            run_get_lock()

        # The lock should still exist
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())

    @freeze_time('2014-02-01')
    def test_lock_before_suppress_acquisition_exceptions(self):
        """
        Tests when a lock already exists. Note that it should not raise an exception since
        suppress_acquisition_exceptions is True.
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')

        @db_mutex('lock_id', suppress_acquisition_exceptions=True)
        def run_get_lock():
            raise NotImplementedError

        # Try to acquire the lock. It should not raise an exception nor acquire the lock
        run_get_lock()

        # The lock should still exist
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())

    @freeze_time('2014-02-01')
    def test_lock_different_id(self):
        """
        Tests that the lock still works even when another lock with a different id exists.
        """
        # Create a lock
        m = DBMutex.objects.create(lock_id='lock_id')

        @db_mutex('lock_id2')
        def run_get_lock2():
            self.assertEqual(DBMutex.objects.count(), 2)
            m2 = DBMutex.objects.get(lock_id='lock_id2')
            self.assertEqual(m2.creation_time, datetime(2014, 2, 1))

        # Try to acquire the lock with a different ID
        run_get_lock2()
        # The original lock should still exist but the other one should be gone
        self.assertTrue(DBMutex.objects.filter(id=m.id).exists())
        self.assertEqual(DBMutex.objects.count(), 1)

    def test_lock_timeout_default(self):
        """
        Tests that the lock timeout works with the default value of 30 minutes.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            orig_lock = DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        @freeze_time('2014-02-01 00:01:00')
        @db_mutex('lock_id')
        def run_get_lock1():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock1()

        # Try to acquire the lock 9 minutes in the future. It should fail
        @freeze_time('2014-02-01 00:09:00')
        @db_mutex('lock_id')
        def run_get_lock2():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock2()

        @freeze_time('2014-02-01 00:30:00')
        @db_mutex('lock_id')
        def run_get_lock3():
            self.assertFalse(DBMutex.objects.filter(id=orig_lock.id).exists())
            self.assertEqual(DBMutex.objects.count(), 1)
            m = DBMutex.objects.get(lock_id='lock_id')
            self.assertEqual(m.creation_time, datetime(2014, 2, 1, 0, 30))

        # Try to acquire the lock 30 minutes in the future. It should pass since the lock timed out
        run_get_lock3()

    @override_settings(DB_MUTEX_TTL_SECONDS=None)
    def test_no_lock_timeout(self):
        """
        Tests that the lock timeout works when None is configured as the timeout.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        @freeze_time('2014-02-01 00:01:00')
        @db_mutex('lock_id')
        def run_get_lock1():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock1()

        # Try to acquire the lock 9 minutes in the future. It should fail
        @freeze_time('2014-02-01 00:09:00')
        @db_mutex('lock_id')
        def run_get_lock2():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock2()

        # Try to acquire the lock 30 minutes in the future. It should fail
        @freeze_time('2014-02-01 00:30:00')
        @db_mutex('lock_id')
        def run_get_lock3():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock3()

        # Try to acquire the lock years in the future. It should fail
        @freeze_time('2016-02-01 00:30:00')
        @db_mutex('lock_id')
        def run_get_lock4():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock4()

    @override_settings(DB_MUTEX_TTL_SECONDS=60 * 60)
    def test_custom_lock_timeout(self):
        """
        Tests that the custom lock timeout works when an hour is configured as the timeout.
        """
        with freeze_time('2014-02-01'):
            # Create a lock
            orig_lock = DBMutex.objects.create(lock_id='lock_id')

        # Try to acquire the lock one minute in the future. It should fail
        @freeze_time('2014-02-01 00:01:00')
        @db_mutex('lock_id')
        def run_get_lock1():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock1()

        # Try to acquire the lock 31 minutes in the future. It should fail
        @freeze_time('2014-02-01 00:31:00')
        @db_mutex('lock_id')
        def run_get_lock2():
            raise NotImplementedError

        with self.assertRaises(DBMutexError):
            run_get_lock2()

        # Try to acquire the lock 60 minutes in the future. It should pass
        @freeze_time('2014-02-01 01:00:00')
        @db_mutex('lock_id')
        def run_get_lock3():
            self.assertFalse(DBMutex.objects.filter(id=orig_lock.id).exists())
            self.assertEqual(DBMutex.objects.count(), 1)
            m = DBMutex.objects.get(lock_id='lock_id')
            self.assertEqual(m.creation_time, datetime(2014, 2, 1, 1))

        run_get_lock3()

    def test_lock_timeout_error(self):
        """
        Tests the case when a lock expires while the context manager is executing.
        """
        @freeze_time('2014-02-01')
        @db_mutex('lock_id')
        def run_get_lock1():
            # Acquire a lock at the given time and release it before it is finished. It
            # should result in an error
            self.assertEqual(DBMutex.objects.count(), 1)
            m = DBMutex.objects.get(lock_id='lock_id')
            self.assertEqual(m.creation_time, datetime(2014, 2, 1))

            # Release the lock before the context manager finishes
            m.delete()

        with self.assertRaises(DBMutexTimeoutError):
            run_get_lock1()
